; PP4X test EPROM v2.0
;
; written by Martin Mansfield 2017-18
;
scratchram	equ	$d000
ramsize		equ	$0800
ramtestval	equ	$55
; ram addresses from endram to ramsize are overlaid by latches
endram		equ	$d7c0		; last usable address+1
ramusable	equ	endram-scratchram

pagedram	equ	$0000
pagesize	equ	$8000
maxpages	equ	256

; 6821 pia address
piabase		equ	$dff8
; reg offsets
; note that cs0 & cs1 are wired incorrectly
piaprega	equ	0
piaddrega	equ	0
piacrega	equ	2
piapregb	equ	1
piaddregb	equ	1
piacregb	equ	3

; default control reg values
prtacrdef	equ	$30
prtbcrdef	equ	$30

; default data direction reg values
prtaddrval	equ	$ff
prtbddrval	equ	$c0

; control reg flags
ddraccess	equ	$00		; data direction register access
prgaccess	equ	$04		; peripheral register access
prtbbeepon	equ	$08|prgaccess
prtbbeepoff	equ	$00|prgaccess

; peripheral reg flags
dispbusy	equ	$10

; 2661B uart address
uart1base	equ	$dff0
uart2base	equ	$dfe8
; reg offsets
uartdata	equ	0
uartstatus	equ	1
uartmode	equ	2
uartcmd		equ	3

; mode & cmd register values for 19200,8,n,1
mod1regval	equ	$4e
mod2regval	equ	$fe
cmdregval	equ	$27

uarttxbit	equ	1

; latch addresses
pclbase		equ	$d7e6
hslbase		equ	$d7e8
rtlbase		equ	$d7ea
dsebase		equ	$d7ec
rbebase		equ	$d7ee
dbebase		equ	$d7f0

; hardware status latch flags
hsldres		equ	$20

; DAC base address
dacbase		equ	$d7f8

; ADL/ADH base addresses
adhbase		equ	$dfd8
adlbase		equ	$dfe0

; display constants
dspposval	equ	$a0
spacechar	equ	$20
crchar		equ	$0d
lfchar		equ	$0a
bytechsize	equ	2	; characters used to display a byte
addrchsize	equ	4	; characters used to display an address
mempgpos	equ	14	; display pos for memory page
memfailpos	equ	2	; display pos for failure info
selprmtpos	equ	10	; display pos for select options
optnumpos	equ	12	; display pos for max option

; keyboard constants
keycnt		equ	20
scanloopval	equ	256	; loop value while no key pressed
; these timings based on 2MHz clock
beeplen		equ	500
chirplen	equ	beeplen/40
beepcnt		equ	200

us100dlval	equ	40	; loop value for 100us delay
sec1dlval	equ	48780	; loop value for 1s delay

;
; temporary buffer
;
tmpbufsize	equ	16
tmpbuf		equ	endram-tmpbufsize

;
; global variables
;
keyrowmskoff	equ	0
keyrowidxoff	equ	1
keycolidxoff	equ	2
keycntoff	equ	3
keyflgoff	equ	4
rowmskoff	equ	5
colidxoff	equ	6
colmskoff	equ	7
colmskoff2	equ	8
scancntoff	equ	9	; word value
scancntoff2	equ	10
;
uarterroff	equ	11	; set if uart missing/faulty

globalvarsize	equ	12

globalvars	equ	tmpbuf-globalvarsize

stackst		equ	globalvars

	org	$e000

piaflp:
	lda	,X		; loop to indicate fault in pia
	bra	piaflp

versionstr	fcc	"PP4X TEST V2-0  "
		fcb	0
dispclr		fcc	"                "
		fcb	0
selprmt		fcc	"0-X ?"
		fcb	0
mainprmt	fcc	"MAIN MENU       "
		fcb	0
stdprmt		fcc	"STD TESTS       "
		fcb	0
dbgprmt		fcc	"DBG TESTS       "
		fcb
memcmdmsg	fcc	"TEST MEM PAGE   "
		fcb	0
sercmdmsg	fcc	"TEST SERIAL     "
		fcb	0
ramcmdmsg	fcc	"DUMP RAM        "
		fcb	0
lpcmdmsg	fcc	"MEM RFSH TIMING "
		fcb	0
wrlpcmdmsg	fcc	"MEM WR/RD TIMING"
		fcb
memfailmsg	fcc	"F"
		fcb	0
uartfailmsg	fcc	"NO UART"
		fcb	0
message		fcc	"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
		fcb	0

hexlookup	fcc	"0123456789ABCDEF"

; menu definitions
menu0		fdb	mainprmt
		fcb	emenu0-cmds0/2
		fcb	0
cmds0
		fdb	stdmenu		; standard tests
		fdb	dbgmenu		; debug tests
emenu0

menu1		fdb	stdprmt
		fcb	emenu1-cmds1/2
		fcb	0
cmds1
		fdb	memtest		; paged memory test
		fdb	txmessage	; serial port test
		fdb	ramdump		; RAM contents dump to serial port
emenu1

menu2		fdb	dbgprmt
		fcb	emenu2-cmds2/2
		fcb	0
cmds2
		fdb	looptest	; mem refresh timing test
		fdb	wrtest		; mem write/read timing test
emenu2

; keycode maps
; hex digits 0 -F
digitmap	fcb	$44,$34,$35,$3a,$24,$25,$2a,$14,$15,$1a,$39,$38,$29,$28,$19,$18,0
; down, up, left, right
cursormap	fcb	$45,$4a,$49,$48,0
; load, input, output, verify, empty, edit, list, hidden1, hidden2, hidden3, csum, set, exit, program
cmdmap		fcb	$11,$12,$13,$17,$16,$21,$22,$23,$27,$26,$41,$42,$43,$46,0

cmdstop		equ	12

romstart:
; hold display in reset
	lda	#hsldres
	sta	hslbase
; configure pia CB2 for buzzer and port B test
	lda	#prtbcrdef|ddraccess
	sta	piabase+piacregb
	lda	#prtbddrval
	sta	piabase+piaddregb
	lda	#prtbcrdef|prgaccess
	sta	piabase+piacregb

; check that pia port B is reading sensible values
	lda	piabase+piapregb
	anda	#$0f			; lower 4 bits should be zero
	beq	piatestend		; yes, so continue

	ldx	#piabase+piapregb
	jmp	piaflp			; indicate pia fault

piatestend:

beeprst:
; 0.5 second beep
	ldx	#beeplen
; set return address to following code
	ldy	#beepend

; sound the buzzer
beep:
bplp:
; turn on buzzer
	lda	#prtbcrdef|prtbbeepon
	sta	piabase+piacregb
	lda	#beepcnt
bplp2:	deca
	bne	bplp2

; turn off buzzer
	lda	#prtbcrdef|prtbbeepoff
	sta	piabase+piacregb
	lda	#beepcnt
bplp3:	deca
	bne	bplp3

	leax	-1,X
	bne	bplp

	jmp	,Y			

beepend:

; simple static RAM test before using stack

scramtest:
	ldu	#scratchram
	ldx	#ramusable
	ldy	#ramtestend

ramtest
rtlp:	ldb	,U		; save original
	lda	#ramtestval
	sta	,U		; store test value
	cmpa	,U
	bne	rterr		; beep loop if error
	coma
	sta	,U		; store aah
	cmpa	,U
	bne	rterr		; beep loop if error
	stb	,U+		; restore contents and inc pointer
	leax	-1,X		; dec count
	bne	rtlp
	bra	rtret		; finished	

rterr:	orcc	#1

rtret:	jmp	,Y

ramtestend:
	bcs	beeprst			; jump to beep again if error

; set up stack pointer
	lds	#stackst
; initialise variables
	jsr	varinit
; initialise uart
	jsr	uartinit
	bcc	uartdone
	com	globalvars+uarterroff	; set error flag
uartdone:
; initialise display and keypad
	jsr	dspinit
; display version
	ldx	#versionstr
	jsr	dspmess

	jsr	s1delay

; display main menu
	ldx	#menu0
	jsr	menudisp
;
; soft reset
;
resettest:
; pia test at startup will fail if key is still pressed
	jsr	keyscanchk		; wait if key still detected
	bcc	resettest		; scan again if returns another keypress
	jmp	[resvec]

; standard test menu
stdmenu:
	ldx	#menu1
	jmp	menudisp

; debug test menu
dbgmenu:
	ldx	#menu2
	jmp	menudisp

;
; test optional memory
;
memtest:
	ldx	#memcmdmsg
	jsr	dspmess

	clra
	ldb	#maxpages
memtestlp:
	sta	rbebase			; select mem page

	pshs	B 
	tfr	A,B
	lda	#mempgpos
	jsr	dsphexbyte		; display memory page
	tfr	B,A
	puls	B 

	jsr	pgramtest
	bcc	memtestnext		; failed ?

memtestfail:
	pshs	U			; save address
	pshs	D
	tfr	U,D
	subd	#pagedram		; convert address to offset
	tfr	D,X
	puls	D

	pshs	X
	jsr	dspclr
	ldx	#memfailmsg
	jsr	dspmess
	puls	X

	pshs	A
	ldb	rbebase			; restore mem page from latch !
	lda	#memfailpos
	jsr	dsphexbyte		; display mem page
	adda	#bytechsize+1
	jsr	dsphexaddr
	adda	#addrchsize+1
	puls	B
	jsr	dsphexbyte
	adda	#bytechsize+1
	puls	X
	ldb	,X
	jsr	dsphexbyte
	jsr	keyscan			; wait for a keypress
	bra	memtestend

memtestnext:
	lda	rbebase			; restore mem page from latch !
	inca
	decb
	bne	memtestlp

memtestend:
	clra
	sta	rbebase			; set to default page
	rts
;
; output a character string to serial port
;
txmessage:
	ldx	#sercmdmsg
	jsr	dspmess

	tst	globalvars+uarterroff	; check uart
	beq	txmessok	; uart present ?
	jsr	dspclr
	ldx	#uartfailmsg	; no
	jsr	dspmess
	jsr	s1delay
	bra	txmessend
txmessok:
	ldx	#message
	clrb
txmesslp:
	jsr	txstring
	decb
	bne	txmesslp
txmessend:
	rts
;
; dump the static RAM contents to serial port
;
ramdump:
	ldx	#ramcmdmsg
	jsr	dspmess

	tst	globalvars+uarterroff	; check uart
	beq	ramdumpok	; uart present ?
	jsr	dspclr
	ldx	#uartfailmsg	; no
	jsr	dspmess
	jsr	s1delay
	bra	ramdumpend
ramdumpok:
	ldu	#scratchram
	ldx	#ramsize
	clrb
ramdumplp:
	lda	,U+			; get contents and increment pointer
	jsr	txhexbyte
	lda	#spacechar
	jsr	txchar
	incb
	leax	-1,X			; decrement count
	tfr	CC,A			; save condition code reg
	bitb	#$0f			; send cr/lf every 16 bytes
	bne	nocrlf
	jsr	txcrlf
nocrlf:
	tfr	A,CC			; restore condition code reg
	bne	ramdumplp
ramdumpend:
	rts

; infinite loop for probing DRAM refresh etc
looptest:
	ldx	#lpcmdmsg
	jsr	dspmess		; display message
looptestlp:
	bra	looptestlp

; infinite loop for probing DRAM refresh etc
wrtest:
	ldx	#wrlpcmdmsg
	jsr	dspmess		; display message
; select memory page zero
	clra
	sta	rbebase		; select mem page

	lda	#$55
	ldx	#pagedram
wrtestlp:
	sta	[,X]
	cmpa	[,X]
	bra	wrtestlp
;
; paged ram test
;
; on return carry clear if passed, carry set if failed
; failing address in u, value stored in a
;
pgramtest:
	pshs	X,Y,B 
	ldu	#pagedram
	ldx	#pagesize
	ldy	#pgramtestret		; provide return address
	jmp	ramtest
pgramtestret:
	puls	X,Y,B 
	rts

;
; display the current menu
;
; pointer to menu definition in x
;
menudisp:
	pshs	X
	ldy	,X++			; get menu title
	ldb	,X++			; get number of commands

	pshs	X
	tfr	Y,X

; display menu title
	jsr	dspmess
; overlay with prompt
	ldx	#selprmt
	lda	#selprmtpos
	jsr	dspmesspos
	puls	X
; customise prompt
	decb				; convert to zero based max value
	lda	#optnumpos
	jsr	dsphexchar		; display max option
	incb

; scan the keyboard
menuscanlp:
	jsr	keyscan

	jsr	getcmdkeycode		; check for exit key
	bcs	menucmdval

	cmpa	#cmdstop		; check for stop key
	bne	menuscanlp		; no
	jsr	keychirp
	bra	menuret			; yes

menucmdval:
	jsr	getdigkeycode
	bcs	menuscanlp

	stb	tmpbuf
	cmpa	tmpbuf			; valid command ?
	bcc	menuscanlp

	jsr	keychirp
	jsr	dspclr			; clear display

; call cmd table entry as subroutine
	lsla				; adjust for pointer to word
	jsr	[A,X]			; call it

menulp:
	puls	X
	bra	menudisp

menuret:
	puls	X
	rts
;
; check if key is digit, cursor or command
;
; key in a
;
; on return key id in a with carry clear if valid
;                           or carry set if not
;
getdigkeycode:
	pshs	B,X
	ldx	#digitmap
	bra	findkey

getcurkeycode:
	pshs	B,X
	ldx	#cursormap
	bra	findkey

getcmdkeycode:
	pshs	B,X
	ldx	#cmdmap

findkey:
	clrb
findkeylp:
	tst	,X			; end of list ?
	bne	findkeycmp		; no
	orcc	#1			; yes, not found, carry set
	bra	findkeyret

findkeycmp:
	cmpa	,X+			; found it ?
	beq	findkeyok		; yes
	incb
	bra	findkeylp
;
findkeyok:
	tfr	B,A			; carry clear
findkeyret:
	puls	B,X
	rts
;
; initialise variables
;
varinit:
;
	clra
	sta	globalvars+uarterroff	

; initialise latch values shadowed in RAM
;
; reading from the latch address will then return
; the last value written to the latch
;
; note: hslbase & dsebase already written to
	sta	pclbase
	sta	rtlbase
	sta	rbebase
	rts
;
; uart subroutines
;
; uartinit
;
; returns with carry clear if ok, set if error
;
uartinit:
	pshs	A,B
; ensure mode register access reset
	ldb	uart1base+uartcmd
; set mode registers 1 & 2
	lda	#mod1regval
	sta	uart1base+uartmode
	lda	#mod2regval
	sta	uart1base+uartmode
; toggle bit 0 of command register
	eorb	#1
	stb	uart1base+uartcmd
; and compare it
	cmpb	uart1base+uartcmd
	beq	uartok
	orcc	#1		; set carry flag
	bra	uartret
; set command register
uartok:
	lda	#cmdregval
	sta	uart1base+uartcmd
uartret:
	puls	A,B
	rts
;
; transmit char
;
txchar:
	pshs	B 	
	ldb	#uarttxbit
; wait for transmitter to empty
txchlp:
	bitb	uart1base+uartstatus
	beq	txchlp
	sta	uart1base+uartdata
	puls	B 
	rts
;
; transmit cr/lf
;
txcrlf:
	pshs	A 
	lda	#crchar
	jsr	txchar
	lda	#lfchar
	jsr	txchar
	puls	A 
	rts
;
; transmit hex byte
;
txhexbyte:
	pshs	A
	lsra
	lsra
	lsra
	lsra
	jsr	txhexchar
	puls	A 
	anda	#$0f
	jsr	txhexchar
	rts
;
; transmit hex char
;
txhexchar:
	pshs	X 
	ldx	#hexlookup
	lda	A,X
	puls	X 
	bra	txchar
;
; transmit a string of chars
;
txstring:
	pshs	A,X 
txstringlp:
	lda	,X+
	beq	txstrret
	jsr	txchar
	bra	txstringlp
txstrret:
	puls	A,X 
	rts
;
; initialise display hardware
;
dspinit:
; configure PIA ports A & B

; port A configured as outputs
	lda	#prtacrdef|ddraccess	; set to data direction register access
	sta	piabase+piacrega
	lda	#prtaddrval
	sta	piabase+piaddrega
	lda	#prtacrdef|prgaccess	; leave for peripheral reg access
	sta	piabase+piacrega

; port B bits 0-5 configured as inputs, 6 & 7 as outputs
	lda	#prtbcrdef|ddraccess	; set to data direction register access
	sta	piabase+piacregb
	lda	#prtbddrval
	sta	piabase+piaddregb
	lda	#prtbcrdef|prgaccess	; leave for peripheral reg access
	sta	piabase+piacregb

; disable reset for display IC
	lda	hslbase			; get current value from shadow RAM
; clear DRES
	anda	#~hsldres
	sta	hslbase
; 100us delay
	jsr	us100delay
; clear display
	jsr	dspclr
; set display duty cycle to 16/32
	lda	#$f0
	jsr	dspsend
; init key scan variables
	clra
	ldy	#globalvars
	sta	keyflgoff,Y
	rts
;
; clear display
;
dspclr:
	pshs	X
	ldx	#dispclr
	jsr	dspmess
	puls	X
	rts
;
; display hexadecimal char
;
; number in b
; zero based start position in a
;
dsphexchar:
	pshs	X 
	pshs	A 
	ldx	#hexlookup
	tfr	B,A
	anda	#$0f
	lda	A,X  
	sta	tmpbuf
	clr	tmpbuf+1
	puls	A 
	ldx	#tmpbuf
	jsr	dspmesspos
	puls	X 
	rts
;
; display hexadecimal address
;
; address in X
; zero based start position in a

dsphexaddr:
	pshs	B
	pshs	X
	puls	B
	jsr	dsphexbyte
	puls	B
	pshs	A
	inca
	inca
	jsr	dsphexbyte
	puls	A
	puls	B
	rts	
;
; display hexadecimal byte
;
; number in b
; zero based start position in a
;
dsphexbyte:
	pshs	X 
	pshs	A 
	ldx	#hexlookup
	tfr	B,A
	lsra
	lsra
	lsra
	lsra
	lda	A,X 
	sta	tmpbuf
	tfr	B,A
	anda	#$0f
	lda	A,X  
	sta	tmpbuf+1
	clr	tmpbuf+2
	puls	A 
	ldx	#tmpbuf
	jsr	dspmesspos
	puls	X 
	rts
;
; display a message on VFD at pos 0
;
; pointer to null terminated message in x
;
dspmess		pshs	A 
		clra
		jsr	dspmesspos
		puls	A 
		rts
;
; display a message on VFD
;
; pointer to null terminated message in x
; zero based start position in a
;
dspmesspos:
	pshs	A 
	deca			; value is 1 less than zero based offset
	anda	#$0f
; set character position
	ora	#dspposval	; build command value
	jsr	dspsend

dspmslp1:
	lda	,X
	beq	dspmsend	; sent all data ?
	anda	#$3f		; clear top 2 bits
	jsr	dspsend
	leax	1,X
	bra	dspmslp1
dspmsend:
	puls	A 
	rts
;
; send byte of data in A reg to 10937 display IC
;
dspsend:
	pshs	B 
; loop until not busy
busylp:
	ldb	piabase+piapregb
	andb	#dispbusy
	bne	busylp

; 100us delay
	jsr	us100delay
; 100us delay
	jsr	us100delay

	sta	dsebase		; write data to DSE latch
	puls	B 
	rts	
;
; scan the keyboard
;
; keyscan - wait until key pressed
;           keycode returned in reg A, top nibble is row, bottom nibble is column
;
; keyscanchk - wait a fixed time for key state change
;            returns with carry set if no state change
;
keyscanchk:
	pshs	X
	ldx	#scanloopval
	bra	keyscancnt
keyscan:
	pshs	X
	ldx	#0		; continue scan until key pressed
keyscancnt:
	pshs	B,Y
	ldy	#globalvars	; init for variable access
	stx	scancntoff,Y	; store scanloop value
	ldx	#piabase
keyscanlp:
	clra
	sta	colidxoff,Y	; set column var to 0
	sta	colmskoff2,Y	; set to 0
	inca
	sta	colmskoff,Y	; set to 1
keyscanlp1:
	lda	colmskoff2,Y	; update bits 8 & 9
	sta	piapregb,X
keyscanlp2:
	lda	colmskoff,Y	; update bits 0-7
	sta	piaprega,X
;
; read lower 4 bits of port b
;
	lda	piapregb,X
	anda	#$0f		; mask out
	sta	rowmskoff,Y	; save row value
	lda	keyflgoff,Y	; current key ?
	beq	keyfirst	; no
	lda	keycolidxoff,Y	; yes
	cmpa	colidxoff,Y	; in this column ?
	bne	keycont		; no, ignore it

; key press already detected in this col
keynext:
	lda	rowmskoff,Y	; key pressed in current col ?
	bne	keydn2		; yes
	bra	keyup1

; no key press already detected
keyfirst:
	lda	rowmskoff,Y	; key pressed in current col ?
	bne	keydn1		; yes
	bra	keycont

; check for multiple keys pressed
keydn2:
	lda	keyrowmskoff,Y
	cmpa	rowmskoff,Y	; the only key pressed ?
	beq	keydn3		; yes
	anda	rowmskoff,Y	; still pressed ?
	bne	keycont		; yes, wait until only key pressed

; detected key up
keyup1:
	lda	keycntoff,Y
	cmpa	#keycnt*2	; has count reached high limit ?
	beq	keyup2		; yes
	inc	keycntoff,Y	; no, increment count
	bra	keycont

keyup2:
	clra
	sta	keyflgoff,Y	; clear flag
	bra	keycont
;
; detected first key down
;
; convert row mask to row index
keydn1:
	ldb	#1
	lda	rowmskoff,Y
keydn1lp:
	lsra
	bcs	keydownchk
	incb
	bra	keydn1lp
;
; check for multiple keys
;
keydownchk:
	bne	keycont		; only one key ?
; set variables
	lda	#keycnt
	sta	keycntoff,Y	; save count for key
	lda	colidxoff,Y
	sta	keycolidxoff,Y	; save column for key
	lda	rowmskoff,Y
	sta	keyrowmskoff,Y	; save row mask for key
	stb	keyrowidxoff,Y	; save row index for key
	lda	#1
	sta	keyflgoff,Y	; set flag
	bra	keycont

;
; detected following key down
;
keydn3:
	lda	keycntoff,Y	; already detected ?
	beq	keycont		; yes, wait for it to be released
	dec	keycntoff,Y	; decrement count for key
	bne	keycont
;
; build key code from row and column
;
keycode:
	lda	keyrowidxoff,Y
	asla
	asla
	asla
	asla			; upper bits are row
	ora	keycolidxoff,Y	; lower bits are column
	inca			; column index is based from 1
	bra	keyscanend	; carry clear
;
; continue scan
;
keycont:
	inc	colidxoff,Y	; inc column var
	lda	colmskoff,Y	; get column mask
	beq	keychk89	; shift bits 8 & 9	
	asl	colmskoff,Y	; scan next column
	bcc	keyscanlp1	; bit 8 set ?
	lda	#$40
	sta	colmskoff2,Y	; init mask
	bra	keyscanlp1
;
keychk89:
	asl	colmskoff2,Y	; scan next column
	bcc	keyscanloop	; first column ?
	tst	keyflgoff,Y	; yes, current key ?
	bne	keyendlp	; yes, continue scan

; check if scan count in use
keycntchk:
	ldd	scancntoff,Y	; no, check if scan count is zero
	beq	keyendlp	; yes, so don't decrement it

; decrement scan count and check for zero
keydeccnt:
	subd	#1		; decrement scan count
	std	scancntoff,Y	; is it zero ?
	bne	keyendlp	; no
	orcc	#1		; set carry and return

keyscanend:
	puls	B,Y
	puls	X
	rts
	
keyendlp:
	jmp	keyscanlp

keyscanloop:
	jmp	keyscanlp1

;
; chirp when key pressed
;
keychirp:
	pshs	A,X,Y 

	ldx	#chirplen
	ldy	#kchpret	; provide return address
	jmp	beep		; call beep code as a subroutine

kchpret:
	puls	A,X,Y 
	rts

; 100us delay
us100delay:
	pshs	A 
	lda	#us100dlval

dllp1:	deca
	bne	dllp1
	puls	A 
	rts

; 1s delay
s1delay:
	pshs	X 
	ldx	#sec1dlval
s1dllp1:
	exg	X,Y		; just to waste some time !
	exg	X,Y
	exg	X,Y
	exg	X,Y
	leax	-1,X
	bne	s1dllp1
	puls	X 
	rts

rti:	rti

	org	$fff0

;	Reserved
	fdb	$ffff
;	Reserved
	fdb	$ffff
;	Reserved
	fdb	$ffff
;	Firq vector
	fdb	rti
;	IRQ vector
	fdb	rti
;	SWI vector
	fdb	rti
;	NMI vector
	fdb	rti
;	Reset vector
resvec:	fdb	romstart
